with Interfaces; use Interfaces;

package body Shardware is

  function TempBytes (Instruction : ByteArr) return ByteArr is 
    TheBytes : ByteArr (1 .. 4);
  begin
    TheBytes (1) := Instruction (7);
    TheBytes (2) := Instruction (8);
    TheBytes (3) := Instruction (9);
    TheBytes (4) := Instruction (10);    
    return TheBytes;
  end TempBytes;

  function UnToSignedInt (Value : Unsigned_32) return Integer is 
    Diff1 : Integer := 0;
    Diff2 : Integer := 0;
  begin
    if Value > 2147483647 then
      Diff1 := Integer(Value - 2147483648);
      Diff2 := Diff1 - 2147483647;
      return Diff2 - 1;
    else
      return Integer(Value);
    end if;
  end UnToSignedInt;

  function BytesToBits (TheBytes : ByteArr) return BitArr is 
    CurByte : Byte := 0; 
    Buff : BitArr (1 .. 32);
    Counter : Integer := 0;
  begin
    for i in 1 .. 32 loop 
      Buff (i) := False;
    end loop;
  for i in 1 .. 4 loop
      CurByte := TheBytes (i);
      for j in reverse 1 .. 8 loop
        if CurByte mod 2 > 0 then
          Buff (j + Counter) := True;
        else
          Buff (j + Counter) := False;
        end if;
        CurByte := CurByte / 2;
      end loop;
      Counter := Counter + 8;
    end loop;
    return Buff;
  end BytesToBits;

  function BitsToBytes (BinNumber : BitArr) return ByteArr is
    ReverseBin : BitArr (1 .. 32);
    I : Integer := 1;
    CurByte : Byte := 0;
    Buff : ByteArr (1 .. 4);
    Counter : Integer := 24;
  begin
    for k in reverse 1 .. 32 loop
      ReverseBin (I) := BinNumber (k); -- TODO: Unbloat this mess somehow?
      I := I + 1;
    end loop;
    for i in 1 .. 4 loop
      CurByte := 0;
      for j in reverse 1 .. 8 loop
        if ReverseBin (j + Counter) then 
          CurByte := CurByte + (2 ** (j - 1));
        end if;
      end loop;
      Counter := Counter - 8;
      Buff (i) := CurByte;
    end loop;
    return Buff;
  end BitsToBytes;

  function MyShiftU (Value : Byte; Amount : Integer) return Unsigned_32 is 
  begin
    return Unsigned_32(Shift_Left(Unsigned_128(Value), Amount));
  end MyShiftU;

  function BytesToU (TheBytes : ByteArr) return Unsigned_32 is
  begin
    return Unsigned_32(MyShiftU(TheBytes (1), 24) 
    + MyShiftU(TheBytes (2), 16)
    + MyShiftU(TheBytes (3), 8) 
    + Unsigned_32(TheBytes (4)));
  end BytesToU;

  function BytesToI (TheBytes : ByteArr) return Integer is
  begin
    return UnToSignedInt(BytesToU(TheBytes));
  end BytesToI;

  function BytesToF (TheBytes : ByteArr) return Float is
    Zero : Float := 0.0;
    Result : Float := 0.0;
    TheBits : BitArr (1 .. 32);
    BitBuff : BitArr (1 .. 32);
    Mantissa : Float := 0.0;
    Exponent : Natural := 0;
    Sign : Float := 1.0;
  begin
    TheBits := BytesToBits(TheBytes);

    if TheBits (1) then
      Sign := -1.0;
    else
      Sign := 1.0;
    end if;

    for i in 1 .. 32 loop 
      if i < 10 then
        null;
      else
        if TheBits (i) then
          Mantissa := Mantissa + 2.0 ** (-1 * (i - 9));
        end if;
      end if;
    end loop;
    
    for i in 1 .. 32 loop 
      if i < 25 then
        BitBuff (i) := False;
      else
        BitBuff (i) := TheBits (i - 23);
      end if;
    end loop;

    Exponent := Natural(BytesToU(BitsToBytes(BitBuff)));

    if Exponent > 0 and Exponent < 254 then
      Mantissa := Mantissa + 1.0;
      Result := Sign * Mantissa * (2.0 ** (Exponent - 127));
    elsif Exponent = 255 and Mantissa = 0.0 then
      Result := Sign * 1.0 / Zero;
    elsif Exponent = 0 then 
      Result := 1.0 + Sign * Mantissa * (2.0 ** (Exponent - 127));
    else 
      Result := 0.0 / Zero;
    end if;

    return Result;
  end BytesToF;

  procedure MovU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    Register := Value;
  end MovU;

  procedure MovI (Register : in out Integer; Value : Integer) is
  begin
    Register := Value;
  end MovI;

  procedure MovF (Register : in out Float; Value : Float) is 
  begin
    Register := Value;
  end MovF;

  procedure AddU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    Register := Register + Value;
  end AddU;

  procedure AddI (Register : in out Integer; Value : Integer) is
  begin
    Register := Register + Value;
  end AddI;

  procedure AddF (Register : in out Float; Value : Float) is 
  begin
    Register := Register + Value;
  end AddF;

  procedure SubU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    if Value > Register then
      raise Constraint_Error with "Unsigned value can't be lower than 0.";
    else
      Register := Register - Value;
    end if;
  end SubU;

  procedure SubI (Register : in out Integer; Value : Integer) is
  begin
    Register := Register - Value;
  end SubI;

  procedure SubF (Register : in out Float; Value : Float) is 
  begin
    Register := Register - Value;
  end SubF;

  procedure MulU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    Register := Register * Value;
  end MulU;

  procedure MulI (Register : in out Integer; Value : Integer) is
  begin
    Register := Register * Value;
  end MulI;

  procedure MulF (Register : in out Float; Value : Float) is 
  begin
    Register := Register * Value;
  end MulF;

  procedure DivU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    if Value = 0 then
      raise Constraint_Error with "Please do not divide by 0.";
    else
      Register := Register / Value;
    end if;
  end DivU;

  procedure DivI (Register : in out Integer; Value : Integer) is
  begin
    if Value = 0 then
      raise Constraint_Error with "Please do not divide by 0.";
    else
      Register := Register / Value;
    end if;
  end DivI;

  procedure DivF (Register : in out Float; Value : Float) is 
  begin
    if Value = 0.0 then
      raise Constraint_Error with "Please do not divide by 0.";
    else
      Register := Register / Value;
    end if;
  end DivF;

end Shardware;
